# distutils: language = c++
cimport cDPg
import numpy as np
from numpy.linalg import norm

cimport numpy as np
from libcpp cimport bool
from cpython cimport array

def coptimum_reparam_N(np.ndarray[double, ndim=1, mode="c"] mf, np.ndarray[double, ndim=1, mode="c"] time,
                      np.ndarray[double, ndim=2, mode="c"] f, 
                      onlyDP=False, rotated=False, isclosed=False, skipm=0,
                      auto=0, w=0.0, lam1=0.0):
    """
    cython interface calculates the warping to align a set of functions f to a single function mf

    :param mf: vector of size N samples of first function
    :param time: vector of size N describing the sample points
    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param onlyDP: use onlyDP (default = False)
    :param rotated: solve for rotation (default = False)
    :param isclosed: is a closed curve (default = False)
    :param skipm: (default 0)
    :param auto: (default 0)
    :param w: barrier weight (default 0.0)
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of f with mf

    """
    cdef int M, N, n1, skipmi, autoi, swap
    cdef double lam, wi
    cdef bool onlyDP1, rotated1, isclosed1
    cdef np.ndarray[double, ndim=1, mode="c"] fopts = np.zeros(5)
    cdef np.ndarray[double, ndim=1, mode="c"] comtime = np.zeros(5)
    lam = lam1
    skipmi = skipm
    autoi = auto
    onlyDP1 = onlyDP
    rotated1 = rotated
    isclosed1 = isclosed
    wi = w
    n1 = 1

    mf = mf / norm(mf)
    M, N = f.shape[0], f.shape[1]

    cdef np.ndarray[double, ndim=1, mode="c"] opti = np.zeros(M+2)
    cdef np.ndarray[double, ndim=1, mode="c"] fi = np.zeros(M)

    gam = np.zeros((M, N))
    for k in xrange(0, N):
        fopts = np.zeros(5)
        comtime = np.zeros(5)
        swap = False
        fi = f[:, k] / norm(f[:, k])
        fi = np.ascontiguousarray(fi)

        cDPg.optimum_reparam(&fi[0], &mf[0], M, n1, wi, onlyDP1, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
        if fopts[0] == 1000:
            cDPg.optimum_reparam(&fi[0], &mf[0], M, n1, wi, True, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
        if swap:
            x = np.arange(1, M+1) / float(M)
            gam[:,k] = np.interp(x,opti[0:M],x)
        else:
            gam[:,k] = opti[0:M]
        
        gam[:,k] = (gam[:,k] - gam[0,k]) / (gam[-1,k] - gam[0,k])

    return gam


def coptimum_reparam_N2(np.ndarray[double, ndim=2, mode="c"] f1, np.ndarray[double, ndim=1, mode="c"] time,
                       np.ndarray[double, ndim=2, mode="c"] f2, 
                      onlyDP=False, rotated=False, isclosed=False, skipm=0,
                      auto=0, w=0.0, lam1=0.0):
    """
    cython interface calculates the warping to align a set of functions f1 to another set of functions f2

    :param f1: numpy ndarray of shape (M,N) of M functions with N samples
    :param time: vector of size N describing the sample points
    :param f2: numpy ndarray of shape (M,N) of M functions with N samples
    :param onlyDP: use onlyDP (default = False)
    :param rotated: solve for rotation (default = False)
    :param isclosed: is a closed curve (default = False)
    :param skipm: (default 0)
    :param auto: (default 0)
    :param w: barrier weight (default 0.0)
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of f with mf

    """
    cdef int M, N, n1, skipmi, autoi
    cdef double lam, wi
    cdef bool onlyDP1, rotated1, isclosed1, swap
    cdef np.ndarray[double, ndim=1, mode="c"] fopts = np.zeros(5)
    cdef np.ndarray[double, ndim=1, mode="c"] comtime = np.zeros(5)
    lam = lam1
    skipmi = skipm
    autoi = auto
    onlyDP1 = onlyDP
    rotated1 = rotated
    isclosed1 = isclosed
    wi = w

    M, N = f1.shape[0], f1.shape[1]
    n1 = 1
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] opti = np.zeros(M+2)
    cdef np.ndarray[double, ndim=1, mode="c"] f1i = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] f2i = np.zeros(M)

    gam = np.zeros((M, N))
    for k in xrange(0, N):
        fopts = np.zeros(5)
        comtime = np.zeros(5)
        swap = False
        f1i = f1[:, k] / norm(f1[:, k])
        f2i = f2[:, k] / norm(f2[:, k])
        f1i = np.ascontiguousarray(f1i)
        f2i = np.ascontiguousarray(f2i)

        cDPg.optimum_reparam(&f1i[0], &f2i[0], M, n1, wi, onlyDP1, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
        if fopts[0] == 1000:
            cDPg.optimum_reparam(&f1i[0], &f2i[0], M, n1, wi, True, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
        if swap:
            x = np.arange(1, M+1) / float(M)
            gami = np.interp(x,opti[0:M],x)
        else:
            gami = opti[0:M]

        gam[:, k] = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam


def coptimum_reparam(np.ndarray[double, ndim=1, mode="c"] f1, np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=1, mode="c"] f2, 
                     onlyDP=False, rotated=False, isclosed=False, skipm=0,
                     auto=0, w=0.0, lam1=0.0):
    """
    cython interface for calculates the warping to align functions f2 to f1

    :param f1: vector of size N samples of first function
    :param time: vector of size N describing the sample points
    :param f2: vector of size N samples of second function
    :param onlyDP: use onlyDP (default = False)
    :param rotated: solve for rotation (default = False)
    :param isclosed: is a closed curve (default = False)
    :param skipm: (default 0)
    :param auto: (default 0)
    :param w: barrier weight (default 0.0)
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align f2 with f1
    """
    cdef int M, n1, skipmi, autoi
    cdef double lam, wi
    cdef bool onlyDP1, rotated1, isclosed1, swap
    cdef np.ndarray[double, ndim=1, mode="c"] fopts = np.zeros(5)
    cdef np.ndarray[double, ndim=1, mode="c"] comtime = np.zeros(5)
    lam = lam1
    skipmi = skipm
    autoi = auto
    onlyDP1 = onlyDP
    rotated1 = rotated
    isclosed1 = isclosed
    swap = False
    wi = w

    M = f1.shape[0]
    n1 = 1

    cdef np.ndarray[double, ndim=1, mode="c"] opti = np.zeros(M+2)
    gam = np.zeros(M)

    cDPg.optimum_reparam(&f1[0], &f2[0], M, n1, wi, onlyDP1, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
    if fopts[0] == 1000:
        cDPg.optimum_reparam(&f1[0], &f2[0], M, n1, wi, True, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
    
    if swap:
        x = np.linspace(0,1,M)
        gam = np.interp(x,opti[0:M],x)
    else:
        gam = opti[0:M]

    gam = (gam - gam[0]) / (gam[-1] - gam[0])

    return gam


def coptimum_reparam_N2_pair(np.ndarray[double, ndim=2, mode="c"] f, np.ndarray[double, ndim=1, mode="c"] time,
                            np.ndarray[double, ndim=2, mode="c"] f1, np.ndarray[double, ndim=2, mode="c"] f2, 
                            onlyDP=False, rotated=False, isclosed=False, skipm=0,
                            auto=0, w=0.0, lam1=0.0):
    """
    cython interface for calculates the warping to align paired function f1 and f2 to f

    :param f: vector of size N samples of first function
    :param time: vector of size N describing the sample points
    :param f1: vector of size N samples of second function
    :param f2: vector of size N samples of second function
    :param onlyDP: use onlyDP (default = False)
    :param rotated: solve for rotation (default = False)
    :param isclosed: is a closed curve (default = False)
    :param skipm: (default 0)
    :param auto: (default 0)
    :param w: barrier weight (default 0.0)
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align f2 with f1
    """
    cdef int M, N, n1, skipmi, autoi
    n1 = 2
    cdef double lam, wi
    cdef bool onlyDP1, rotated1, isclosed1, swap
    cdef np.ndarray[double, ndim=1, mode="c"] fopts = np.zeros(5)
    cdef np.ndarray[double, ndim=1, mode="c"] comtime = np.zeros(5)
    lam = lam1
    skipmi = skipm
    autoi = auto
    onlyDP1 = onlyDP
    rotated1 = rotated
    isclosed1 = isclosed
    wi = w
    M, N = f1.shape[0], f1.shape[1]
    lam = lam1

    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] opti = np.zeros(M+2)
    cdef np.ndarray[double, ndim=1, mode="c"] f1i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] f2i = np.zeros(M * n1)

    gam = np.zeros((M, N))
    for k in xrange(0, N):
        fopts = np.zeros(5)
        comtime = np.zeros(5)
        swap = False
        f1i = f.reshape(M*n1)
        f2tmp = np.column_stack((f1[:, k], f2[:, k]))
        f2i = f2tmp.reshape(M*n1)

        f1i = np.ascontiguousarray(f1i)
        f2i = np.ascontiguousarray(f2i)

        cDPg.optimum_reparam(&f1i[0], &f2i[0], M, n1, wi, onlyDP1, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
        if fopts[0] == 1000:
            cDPg.optimum_reparam(&f1i[0], &f2i[0], M, n1, wi, True, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
        if swap:
            x = np.arange(1, M+1) / float(M)
            gami = np.interp(x,opti[0:M],x)
        else:
            gami = opti[0:M]

        gam[:, k] = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam


def coptimum_reparam_pair_f(np.ndarray[double, ndim=2, mode="c"] f1, np.ndarray[double, ndim=1, mode="c"] time,
                          np.ndarray[double, ndim=2, mode="c"] f2, 
                          onlyDP=False, rotated=False, isclosed=False, skipm=0,
                          auto=0, w=0.0, lam1=0.0):
    """
    cython interface for calculates the warping to align paired function f2 to f1

    :param f1: vector of size N samples of first function
    :param time: vector of size N describing the sample points
    :param f2: vector of size N samples of second function
    :param onlyDP: use onlyDP (default = False)
    :param rotated: solve for rotation (default = False)
    :param isclosed: is a closed curve (default = False)
    :param skipm: (default 0)
    :param auto: (default 0)
    :param w: barrier weight (default 0.0)
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align f2 with f1
    """
    cdef int M, N, skipmi, autoi
    cdef double lam, wi
    cdef bool onlyDP1, rotated1, isclosed1, swap
    cdef np.ndarray[double, ndim=1, mode="c"] fopts = np.zeros(5)
    cdef np.ndarray[double, ndim=1, mode="c"] comtime = np.zeros(5)
    M, N = f1.shape[0], f1.shape[1]
    lam = lam1
    skipmi = skipm
    autoi = auto
    onlyDP1 = onlyDP
    rotated1 = rotated
    isclosed1 = isclosed
    wi = w
    swap = False

    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] opti = np.zeros(M+2)
    cdef np.ndarray[double, ndim=1, mode="c"] f1i = np.zeros(M * N)
    cdef np.ndarray[double, ndim=1, mode="c"] f2i = np.zeros(M * N)

    sizes = np.zeros(1, dtype=np.int32)
    f1i = f1.reshape(M*N)
    f2i = f2.reshape(M*N)

    f1i = np.ascontiguousarray(f1i)
    f2i = np.ascontiguousarray(f2i)

    cDPg.optimum_reparam(&f2i[0], &f2i[0], M, N, wi, onlyDP1, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
    if fopts[0] == 1000:
        cDPg.optimum_reparam(&f2i[0], &f2i[0], M, N, wi, True, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
    
    if swap:
        x = np.arange(1, M+1) / float(M)
        gami = np.interp(x,opti[0:M],x)
    else:
        gami = opti[0:M]

    gam = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam


def coptimum_reparam_curve_f(np.ndarray[double, ndim=2, mode="c"] f1, np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=2, mode="c"] f2, 
                      onlyDP=False, rotated=False, isclosed=False, skipm=0,
                      auto=0, w=0.0, lam1=0.0):
    """
    cython interface for calculates the warping to align curve f2 to f1

    :param f1: matrix of size nxN samples of first SRVF
    :param time: vector of size N describing the sample points
    :param f2: matrix of size nxN samples of second SRVF
    :param onlyDP: use onlyDP (default = False)
    :param rotated: solve for rotation (default = False)
    :param isclosed: is a closed curve (default = False)
    :param skipm: (default 0)
    :param auto: (default 0)
    :param w: barrier weight (default 0.0)
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align f2 with f1
    """
    cdef int M, n1, skipmi, autoi
    cdef double lam, wi
    cdef bool onlyDP1, rotated1, isclosed1, swap
    cdef np.ndarray[double, ndim=1, mode="c"] fopts = np.zeros(5)
    cdef np.ndarray[double, ndim=1, mode="c"] comtime = np.zeros(5)
    n1 = f1.shape[0]
    M = f1.shape[1]
    lam = lam1
    skipmi = skipm
    autoi = auto
    onlyDP1 = onlyDP
    rotated1 = rotated
    isclosed1 = isclosed
    wi = w
    swap = False

    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] opti = np.zeros(M+2)
    cdef np.ndarray[double, ndim=1, mode="c"] f1i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] f2i = np.zeros(M * n1)

    f1i = f1.reshape(M*n1, order='F')
    f2i = f2.reshape(M*n1, order='F')

    f1i = np.ascontiguousarray(f1i)
    f2i = np.ascontiguousarray(f2i)

    cDPg.optimum_reparam(&f1i[0], &f2i[0], M, n1, wi, onlyDP1, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
        
    if fopts[0] == 1000:
        cDPg.optimum_reparam(&f1i[0], &f2i[0], M, n1, wi, True, rotated1, isclosed1, skipmi, autoi, &opti[0], swap, &fopts[0], &comtime[0])
    
    if swap:
        x = np.arange(1, M+1) / float(M)
        gami = np.interp(x,opti[0:M],x)
    else:
        gami = opti[0:M]

    gam = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam



"""
statistic calculation for SRVF (curves) open and closed using Karcher
Mean and Variance

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
from numpy import zeros, sqrt, fabs, cos, sin, tile, vstack, empty, cov, inf, mean, arange
from numpy.linalg import svd
from numpy.random import randn
import fdasrsf.curve_functions as cf
import fdasrsf.utility_functions as uf
import fdasrsf.plot_style as plot
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
import collections

class fdacurve:
    """
    This class provides alignment methods for open and closed curves using the SRVF framework

    Usage: obj = fdacurve(beta, mode, N, scale)
    :param beta: numpy ndarray of shape (n, M, N) describing N curves
    in R^M
    :param mode: Open ('O') or closed curve ('C') (default 'O')
    :param N: resample curve to N points
    :param scale: scale curve to length 1 (true/false)
    :param q:        (n,T,K) matrix defining n dimensional srvf on T samples with K srvfs
    :param betan:     aligned curves
    :param qn:        aligned srvfs
    :param basis:     calculated basis
    :param beta_mean: karcher mean curve
    :param q_mean:    karcher mean srvf
    :param gams:      warping functions
    :param v:         shooting vectors
    :param C:         karcher covariance
    :param s:         pca singular values
    :param U:         pca singular vectors
    :param coef:      pca coefficients
    :param qun:       cost function
    :param samples:   random samples
    :param gamr:      random warping functions
    :param cent:      center
    :param scale:     scale
    :param E:         energy

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  26-Aug-2020
    """

    def __init__(self, beta, mode='O', N=200, scale=True):
        """
        fdacurve Construct an instance of this class
        :param beta: (n,T,K) matrix defining n dimensional curve on T samples with K curves
        :param mode:  Open ('O') or closed curve ('C') (default 'O')
        :param N: resample curve to N points
        :param scale: scale curve to length 1 (true/false)
        """
        self.mode = mode
        self.scale = scale

        K = beta.shape[2]
        n = beta.shape[0]
        q = zeros((n,N,K))
        beta1 = zeros((n,N,K))
        cent1 = zeros((n,K))
        for ii in range(0,K):
            beta1[:,:,ii] = cf.resamplecurve(beta[:,:,ii],N,mode)
            a = -cf.calculatecentroid(beta1[:,:,ii])
            beta1[:,:,ii] += tile(a, (N,1)).T
            q[:,:,ii] = cf.curve_to_q(beta1[:,:,ii], self.scale, self.mode)
            cent1[:,ii] = -a
        
        self.q = q
        self.beta = beta
        self.cent = cent1


    def karcher_mean(self, parallel=False, cores=-1):
        """
        This calculates the mean of a set of curves
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        """
        n, T, N = self.beta.shape

        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        # Initialize mu as one of the shapes
        mu = self.q[:, :, 0]
        betamean = self.beta[:,:,0]
        itr = 0

        gamma = zeros((T,N))
        maxit = 20

        sumd = zeros(maxit+1)
        v = zeros((n, T, N))
        normvbar = zeros(maxit+1)

        delta = 0.5
        tolv = 1e-4
        told = 5*1e-3

        print("Computing Karcher Mean of %d curves in SRVF space.." % N)
        while itr < maxit:
            print("updating step: %d" % (itr+1))

            if iter == maxit:
                print("maximal number of iterations reached")
            
            mu = mu / sqrt(cf.innerprod_q2(mu, mu))
            if mode == 1:
                self.basis = cf.find_basis_normal(mu)
            else:
                self.basis = []

            sumv = zeros((n, T))
            sumd[0] = inf
            sumd[itr+1] = 0
            out = Parallel(n_jobs=-1)(delayed(karcher_calc)(self.beta[:, :, n],
                                    self.q[:, :, n], betamean, mu, self.basis, mode) for n in range(N))
            v = zeros((n, T, N))
            for i in range(0, N):
                v[:, :, i] = out[i][0]
                sumd[itr+1] = sumd[itr+1] + out[i][1]**2

            sumv = v.sum(axis=2)

            # Compute average direction of tangent vectors v_i
            vbar = sumv/float(N)

            normvbar[itr] = sqrt(cf.innerprod_q2(vbar, vbar))
            normv = normvbar[itr]

            if normv > tolv and fabs(sumd[itr+1]-sumd[itr]) > told:
                # Update mu in direction of vbar
                mu = cos(delta*normvbar[itr])*mu + sin(delta*normvbar[itr]) * vbar/normvbar[itr]

                if mode == 1:
                    mu = cf.project_curve(mu)

                x = cf.q_to_curve(mu)
                a = -1*cf.calculatecentroid(x)
                betamean = x + tile(a, [T, 1]).T
            else:
                break

            itr += 1
        
        self.q_mean = mu
        self.beta_mean = betamean
        self.v = v
        self.qun = sumd[0:(itr+1)]
        self.E = normvbar[0:(itr+1)]

        return


    def srvf_align(self):
        """
        This calculates the mean of a set of curves and aligns them
        """
        n, T, N = self.beta.shape
        # find mean
        if not hasattr(self, 'beta_mean'):
            self.karcher_mean()

        self.qn = zeros((n, T, N))
        self.betan = zeros((n, T, N))
        self.gams = zeros((T,N))
        C = zeros((T,N))
        centroid2 = cf.calculatecentroid(self.beta_mean)
        self.beta_mean = self.beta_mean - tile(centroid2, [T, 1]).T
        q_mu = cf.curve_to_q(self.beta_mean)
        # align to mean
        for ii in range(0, N):
            beta1 = self.beta[:, :, ii]

            # Iteratively optimize over SO(n) x Gamma
            for i in range(0, 1):
                # Optimize over SO(n)
                beta1, O_hat, tau = cf.find_rotation_and_seed_coord(self.beta_mean,
                                                                    beta1)
                q1 = cf.curve_to_q(beta1)

                # Optimize over Gamma
                gam = cf.optimum_reparam_curve(q1, q_mu, 0.0)
                gamI = uf.invertGamma(gam)
                # Applying optimal re-parameterization to the second curve
                beta1 = cf.group_action_by_gamma_coord(beta1, gamI)

            # Optimize over SO(n)
            beta1, O_hat, tau = cf.find_rotation_and_seed_coord(self.beta_mean, beta1)
            self.gams[:,ii] = gamI
            self.qn[:, :, ii] = cf.curve_to_q(beta1)
            self.betan[:, :, ii] = beta1

        return


    def karcher_cov(self):
        """
        This calculates the mean of a set of curves

        """
        if not hasattr(self, 'beta_mean'):
            self.karcher_mean()
        M,N,K = self.v.shape
        tmpv = zeros((M*N,K))
        for i in range(0,K):
            tmp = self.v[:,:,i]
            tmpv[:,i] = tmp.flatten()

        self.C = cov(tmpv)

        return


    def shape_pca(self, no=10):
        """
        Computes principal direction of variation specified by no. N is
        Number of shapes away from mean. Creates 2*N+1 shape sequence

        :param no: number of direction (default 3)
        """
        if not hasattr(self, 'C'):
            self.karcher_cov()

        U1, s, V = svd(self.C)
        self.U = U1[:,0:no]
        self.s = s[0:no]

        # express shapes as coefficients
        K = self.beta.shape[2]
        VM = mean(self.v,2)
        x = zeros((no,K))
        for ii in range(0,K):
            tmpv = self.v[:,:,ii]
            Utmp = self.U.T
            x[:,ii] = Utmp.dot((tmpv.flatten()- VM.flatten()))
        
        self.coef = x

        return


    def sample_shapes(self, no=3, numSamp=10):
        """
        Computes sample shapes from mean and covariance

        :param no: number of direction (default 3)
        :param numSamp: number of samples (default 10)
        """
        n, T = self.q_mean.shape
        modes = ['O', 'C']
        mode = [i for i, x in enumerate(modes) if x == self.mode]
        if len(mode) == 0:
            mode = 0
        else:
            mode = mode[0]

        U, s, V = svd(self.C)

        if mode == 0:
            N = 2
        else:
            N = 10

        epsilon = 1./(N-1)

        samples = empty(numSamp, dtype=object)
        for i in range(0, numSamp):
            v = zeros((2, T))
            for m in range(0, no):
                v = v + randn()*sqrt(s[m])*vstack((U[0:T, m], U[T:2*T, m]))

            q1 = self.q_mean
            for j in range(0, N-1):
                normv = sqrt(cf.innerprod_q2(v, v))

                if normv < 1e-4:
                    q2 = self.q_mean
                else:
                    q2 = cos(epsilon*normv)*q1+sin(epsilon*normv)*v/normv
                    if mode == 1:
                        q2 = cf.project_curve(q2)

                # Parallel translate tangent vector
                basis2 = cf.find_basis_normal(q2)
                v = cf.parallel_translate(v, q1, q2, basis2, mode)

                q1 = q2

            samples[i] = cf.q_to_curve(q2)

        self.samples = samples
        return
    

    def plot(self):
        """
        plot curve mean results
        """
        fig, ax = plt.subplots()
        n,T,K = self.beta.shape
        for ii in range(0,K):
            ax.plot(self.beta[0,:,ii],self.beta[1,:,ii])
        plt.title('Curves')
        ax.set_aspect('equal')
        plt.axis('off')
        plt.gca().invert_yaxis()
        
        if hasattr(self,'gams'):
            M = self.gams.shape[0]
            fig, ax = plot.f_plot(arange(0, M) / float(M - 1), self.gams,
                                    title="Warping Functions")

        if hasattr(self,'beta_mean'):
            fig, ax = plt.subplots()
            ax.plot(self.beta_mean[0,:],self.beta_mean[1,:])
            plt.title('Karcher Mean')
            ax.set_aspect('equal')
            plt.axis('off')
            plt.gca().invert_yaxis()
        
        plt.show()
    

    def plot_pca(self):

        if not hasattr(self,'s'):
            raise NameError('Calculate PCA')
        
        fig, ax = plt.subplots()
        ax.plot(self.s)
        plt.title('Singular Values')

        # plot principal modes of variability
        VM = mean(self.v,2)
        VM = VM.flatten()
        for j in range(0,4):
            fig, ax = plt.subplots()
            for i in range(1,11):
                tmp = VM + 0.5*(i-5)*sqrt(self.s[j])*self.U[:,j]
                m,n = self.q_mean.shape
                v1 = tmp.reshape(m,n)
                q2n = cf.elastic_shooting(self.q_mean,v1)

                p = cf.q_to_curve(q2n)
                if i == 5:
                    ax.plot(0.2*i + p[0,:], p[1,:], 'k', linewidth=2)
                else:
                    ax.plot(0.2*i + p[0,:], p[1,:], linewidth=2)
                
            ax.set_aspect('equal')
            plt.axis('off')
            plt.title('PD %d' % (j+1))
        
        plt.show()


def karcher_calc(beta, q, betamean, mu, basis, mode):
    # Compute shooting vector from mu to q_i
    w, d = cf.inverse_exp_coord(betamean, beta)

    # Project to tangent space of manifold to obtain v_i
    if mode == 0:
        v = w
    else:
        v = cf.project_tangent(w, q, basis)

    return(v, d)



"""
Warping Invariant Regression using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf.utility_functions as uf
from scipy import dot
from scipy.optimize import fmin_l_bfgs_b
from scipy.integrate import trapz
from scipy.linalg import inv, norm
from patsy import bs
from joblib import Parallel, delayed
import mlogit_warp as mw
import collections


def elastic_regression(f, y, time, B=None, lam=0, df=20, max_itr=20,
                       cores=-1, smooth=False):
    """
    This function identifies a regression model with phase-variability
    using elastic methods

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy array of N responses
    :param time: vector of size M describing the sample points
    :param B: optional matrix describing Basis elements
    :param lam: regularization parameter (default 0)
    :param df: number of degrees of freedom B-spline (default 20)
    :param max_itr: maximum number of iterations (default 20)
    :param cores: number of cores for parallel processing (default all)
    :type f: np.ndarray
    :type time: np.ndarray

    :rtype: tuple of numpy array
    :return alpha: alpha parameter of model
    :return beta: beta(t) of model
    :return fn: aligned functions - numpy ndarray of shape (M,N) of M
    functions with N samples
    :return qn: aligned srvfs - similar structure to fn
    :return gamma: calculated warping functions
    :return q: original training SRSFs
    :return B: basis matrix
    :return b: basis coefficients
    :return SSE: sum of squared error

    """
    
    M = f.shape[0]
    N = f.shape[1]

    if M > 500:
        parallel = True
    elif N > 100:
        parallel = True
    else:
        parallel = False

    binsize = np.diff(time)
    binsize = binsize.mean()

    # Create B-Spline Basis if none provided
    if B is None:
        B = bs(time, df=df, degree=4, include_intercept=True)
    Nb = B.shape[1]

    # second derivative for regularization
    Bdiff = np.zeros((M, Nb))
    for ii in range(0, Nb):
        Bdiff[:, ii] = np.gradient(np.gradient(B[:, ii], binsize), binsize)

    q = uf.f_to_srsf(f, time, smooth)

    gamma = np.tile(np.linspace(0, 1, M), (N, 1))
    gamma = gamma.transpose()

    itr = 1
    SSE = np.zeros(max_itr)
    while itr <= max_itr:
        print("Iteration: %d" % itr)
        # align data
        fn = np.zeros((M, N))
        qn = np.zeros((M, N))
        for ii in range(0, N):
            fn[:, ii] = np.interp((time[-1] - time[0]) * gamma[:, ii] +
                                  time[0], time, f[:, ii])
            qn[:, ii] = uf.warp_q_gamma(time, q[:, ii], gamma[:, ii])

        # OLS using basis
        Phi = np.ones((N, Nb+1))
        for ii in range(0, N):
            for jj in range(1, Nb+1):
                Phi[ii, jj] = trapz(qn[:, ii] * B[:, jj-1], time)

        R = np.zeros((Nb+1, Nb+1))
        for ii in range(1, Nb+1):
            for jj in range(1, Nb+1):
                R[ii, jj] = trapz(Bdiff[:, ii-1] * Bdiff[:, jj-1], time)

        xx = dot(Phi.T, Phi)
        inv_xx = inv(xx + lam * R)
        xy = dot(Phi.T, y)
        b = dot(inv_xx, xy)

        alpha = b[0]
        beta = B.dot(b[1:Nb+1])
        beta = beta.reshape(M)

        # compute the SSE
        int_X = np.zeros(N)
        for ii in range(0, N):
            int_X[ii] = trapz(qn[:, ii] * beta, time)

        SSE[itr - 1] = sum((y.reshape(N) - alpha - int_X) ** 2)

        # find gamma
        gamma_new = np.zeros((M, N))
        if parallel:
            out = Parallel(n_jobs=cores)(delayed(regression_warp)(beta,
                                         time, q[:, n], y[n], alpha) for n in range(N))
            gamma_new = np.array(out)
            gamma_new = gamma_new.transpose()
        else:
            for ii in range(0, N):
                gamma_new[:, ii] = regression_warp(beta, time, q[:, ii],
                                                   y[ii], alpha)

        if norm(gamma - gamma_new) < 1e-5:
            break
        else:
            gamma = gamma_new

        itr += 1

    # Last Step with centering of gam
    gamI = uf.SqrtMeanInverse(gamma_new)
    gamI_dev = np.gradient(gamI, 1 / float(M - 1))
    beta = np.interp((time[-1] - time[0]) * gamI + time[0], time,
                     beta) * np.sqrt(gamI_dev)

    for ii in range(0, N):
        qn[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
                              time, qn[:, ii]) * np.sqrt(gamI_dev)
        fn[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
                              time, fn[:, ii])
        gamma[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
                                 time, gamma_new[:, ii])

    model = collections.namedtuple('model', ['alpha', 'beta', 'fn',
                                   'qn', 'gamma', 'q', 'B', 'b',
                                   'SSE', 'type'])
    out = model(alpha, beta, fn, qn, gamma, q, B, b[1:-1], SSE[0:itr],
                'linear')
    return out


def elastic_logistic(f, y, time, B=None, df=20, max_itr=20, cores=-1,
                     smooth=False):
    """
    This function identifies a logistic regression model with
    phase-variability using elastic methods

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy array of labels (1/-1)
    :param time: vector of size M describing the sample points
    :param B: optional matrix describing Basis elements
    :param df: number of degrees of freedom B-spline (default 20)
    :param max_itr: maximum number of iterations (default 20)
    :param cores: number of cores for parallel processing (default all)
    :type f: np.ndarray
    :type time: np.ndarray

    :rtype: tuple of numpy array
    :return alpha: alpha parameter of model
    :return beta: beta(t) of model
    :return fn: aligned functions - numpy ndarray of shape (M,N) of M
    functions with N samples
    :return qn: aligned srvfs - similar structure to fn
    :return gamma: calculated warping functions
    :return q: original training SRSFs
    :return B: basis matrix
    :return b: basis coefficients
    :return Loss: logistic loss

    """
    M = f.shape[0]
    N = f.shape[1]

    if M > 500:
        parallel = True
    elif N > 100:
        parallel = True
    else:
        parallel = False

    binsize = np.diff(time)
    binsize = binsize.mean()

    # Create B-Spline Basis if none provided
    if B is None:
        B = bs(time, df=df, degree=4, include_intercept=True)
    Nb = B.shape[1]

    q = uf.f_to_srsf(f, time, smooth)

    gamma = np.tile(np.linspace(0, 1, M), (N, 1))
    gamma = gamma.transpose()

    itr = 1
    LL = np.zeros(max_itr)
    while itr <= max_itr:
        print("Iteration: %d" % itr)
        # align data
        fn = np.zeros((M, N))
        qn = np.zeros((M, N))
        for ii in range(0, N):
            fn[:, ii] = np.interp((time[-1] - time[0]) * gamma[:, ii] +
                                  time[0], time, f[:, ii])
            qn[:, ii] = uf.warp_q_gamma(time, q[:, ii], gamma[:, ii])

        Phi = np.ones((N, Nb+1))
        for ii in range(0, N):
            for jj in range(1, Nb+1):
                Phi[ii, jj] = trapz(qn[:, ii] * B[:, jj-1], time)

        # Find alpha and beta using l_bfgs
        b0 = np.zeros(Nb+1)
        out = fmin_l_bfgs_b(logit_loss, b0, fprime=logit_gradient,
                            args=(Phi, y), pgtol=1e-10, maxiter=200,
                            maxfun=250, factr=1e-30)
        b = out[0]
        alpha = b[0]
        beta = B.dot(b[1:Nb+1])
        beta = beta.reshape(M)

        # compute the logistic loss
        LL[itr - 1] = logit_loss(b, Phi, y)

        # find gamma
        gamma_new = np.zeros((M, N))
        if parallel:
            out = Parallel(n_jobs=cores)(delayed(logistic_warp)(beta, time,
                                         q[:, n], y[n]) for n in range(N))
            gamma_new = np.array(out)
            gamma_new = gamma_new.transpose()
        else:
            for ii in range(0, N):
                gamma_new[:, ii] = logistic_warp(beta, time, q[:, ii], y[ii])

        if norm(gamma - gamma_new) < 1e-5:
            break
        else:
            gamma = gamma_new

        itr += 1

    # Last Step with centering of gam
    gamma = gamma_new
    # gamI = uf.SqrtMeanInverse(gamma)
    # gamI_dev = np.gradient(gamI, 1 / float(M - 1))
    # beta = np.interp((time[-1] - time[0]) * gamI + time[0], time,
    #                  beta) * np.sqrt(gamI_dev)

    # for ii in range(0, N):
    #     qn[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
    #                           time, qn[:, ii]) * np.sqrt(gamI_dev)
    #     fn[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
    #                           time, fn[:, ii])
    #     gamma[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
    #                              time, gamma[:, ii])

    model = collections.namedtuple('model', ['alpha', 'beta', 'fn',
                                   'qn', 'gamma', 'q', 'B', 'b',
                                   'Loss', 'type'])
    out = model(alpha, beta, fn, qn, gamma, q, B, b[1:-1], LL[0:itr],
                'logistic')
    return out


def elastic_mlogistic(f, y, time, B=None, df=20, max_itr=20, cores=-1,
                      delta=.01, parallel=True, smooth=False):
    """
    This function identifies a multinomial logistic regression model with
    phase-variability using elastic methods

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy array of labels {1,2,...,m} for m classes
    :param time: vector of size M describing the sample points
    :param B: optional matrix describing Basis elements
    :param df: number of degrees of freedom B-spline (default 20)
    :param max_itr: maximum number of iterations (default 20)
    :param cores: number of cores for parallel processing (default all)
    :type f: np.ndarray
    :type time: np.ndarray

    :rtype: tuple of numpy array
    :return alpha: alpha parameter of model
    :return beta: beta(t) of model
    :return fn: aligned functions - numpy ndarray of shape (M,N) of N
    functions with M samples
    :return qn: aligned srvfs - similar structure to fn
    :return gamma: calculated warping functions
    :return q: original training SRSFs
    :return B: basis matrix
    :return b: basis coefficients
    :return Loss: logistic loss

    """
    M = f.shape[0]
    N = f.shape[1]
    # Code labels
    m = y.max()
    Y = np.zeros((N, m), dtype=int)
    for ii in range(0, N):
        Y[ii, y[ii]-1] = 1

    binsize = np.diff(time)
    binsize = binsize.mean()

    # Create B-Spline Basis if none provided
    if B is None:
        B = bs(time, df=df, degree=4, include_intercept=True)
    Nb = B.shape[1]

    q = uf.f_to_srsf(f, time, smooth)

    gamma = np.tile(np.linspace(0, 1, M), (N, 1))
    gamma = gamma.transpose()

    itr = 1
    LL = np.zeros(max_itr)
    while itr <= max_itr:
        print("Iteration: %d" % itr)
        # align data
        fn = np.zeros((M, N))
        qn = np.zeros((M, N))
        for ii in range(0, N):
            fn[:, ii] = np.interp((time[-1] - time[0]) * gamma[:, ii] +
                                  time[0], time, f[:, ii])
            qn[:, ii] = uf.warp_q_gamma(time, q[:, ii], gamma[:, ii])

        Phi = np.ones((N, Nb+1))
        for ii in range(0, N):
            for jj in range(1, Nb+1):
                Phi[ii, jj] = trapz(qn[:, ii] * B[:, jj-1], time)

        # Find alpha and beta using l_bfgs
        b0 = np.zeros(m * (Nb+1))
        out = fmin_l_bfgs_b(mlogit_loss, b0, fprime=mlogit_gradient,
                            args=(Phi, Y), pgtol=1e-10, maxiter=200,
                            maxfun=250, factr=1e-30)
        b = out[0]
        B0 = b.reshape(Nb+1, m)
        alpha = B0[0, :]
        beta = np.zeros((M, m))
        for i in range(0, m):
            beta[:, i] = B.dot(B0[1:Nb+1, i])

        # compute the logistic loss
        LL[itr - 1] = mlogit_loss(b, Phi, Y)

        # find gamma
        gamma_new = np.zeros((M, N))
        if parallel:
            out = Parallel(n_jobs=cores)(delayed(mlogit_warp_grad)(alpha, beta,
                                         time, q[:, n], Y[n, :], delta=delta) for n in range(N))
            gamma_new = np.array(out)
            gamma_new = gamma_new.transpose()
        else:
            for ii in range(0, N):
                gamma_new[:, ii] = mlogit_warp_grad(alpha, beta, time,
                                                    q[:, ii], Y[ii, :], delta=delta)

        if norm(gamma - gamma_new) < 1e-5:
            break
        else:
            gamma = gamma_new

        itr += 1

    # Last Step with centering of gam
    gamma = gamma_new
    # gamI = uf.SqrtMeanInverse(gamma)
    # gamI_dev = np.gradient(gamI, 1 / float(M - 1))
    # beta = np.interp((time[-1] - time[0]) * gamI + time[0], time,
    #                  beta) * np.sqrt(gamI_dev)

    # for ii in range(0, N):
    #     qn[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
    #                           time, qn[:, ii]) * np.sqrt(gamI_dev)
    #     fn[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
    #                           time, fn[:, ii])
    #     gamma[:, ii] = np.interp((time[-1] - time[0]) * gamI + time[0],
    #                              time, gamma[:, ii])

    model = collections.namedtuple('model', ['alpha', 'beta', 'fn',
                                   'qn', 'gamma', 'q', 'B', 'b',
                                   'Loss', 'n_classes', 'type'])
    out = model(alpha, beta, fn, qn, gamma, q, B, b[1:-1], LL[0:itr],
                m, 'mlogistic')
    return out


def elastic_prediction(f, time, model, y=None, smooth=False):
    """
    This function performs prediction from an elastic regression model
    with phase-variability

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size M describing the sample points
    :param model: identified model from elastic_regression
    :param y: truth, optional used to calculate SSE

    :rtype: tuple of numpy array
    :return alpha: alpha parameter of model
    :return beta: beta(t) of model
    :return fn: aligned functions - numpy ndarray of shape (M,N) of N
    functions with M samples
    :return qn: aligned srvfs - similar structure to fn
    :return gamma: calculated warping functions
    :return q: original training SRSFs
    :return B: basis matrix
    :return b: basis coefficients
    :return SSE: sum of squared error

    """
    q = uf.f_to_srsf(f, time, smooth)
    n = q.shape[1]

    if model.type == 'linear' or model.type == 'logistic':
        y_pred = np.zeros(n)
    elif model.type == 'mlogistic':
        m = model.n_classes
        y_pred = np.zeros((n, m))

    for ii in range(0, n):
        diff = model.q - q[:, ii][:, np.newaxis]
        dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1. / 2)
        q_tmp = uf.warp_q_gamma(time, q[:, ii],
                                model.gamma[:, dist.argmin()])
        if model.type == 'linear':
            y_pred[ii] = model.alpha + trapz(q_tmp * model.beta, time)
        elif model.type == 'logistic':
            y_pred[ii] = model.alpha + trapz(q_tmp * model.beta, time)
        elif model.type == 'mlogistic':
            for jj in range(0, m):
                y_pred[ii, jj] = model.alpha[jj] + trapz(q_tmp * model.beta[:, jj], time)

    if y is None:
        if model.type == 'linear':
            SSE = None
        elif model.type == 'logistic':
            y_pred = phi(y_pred)
            y_labels = np.ones(n)
            y_labels[y_pred < 0.5] = -1
            PC = None
        elif model.type == 'mlogistic':
            y_pred = phi(y_pred.ravel())
            y_pred = y_pred.reshape(n, m)
            y_labels = y_pred.argmax(axis=1)+1
            PC = None
    else:
        if model.type == 'linear':
            SSE = sum((y - y_pred) ** 2)
        elif model.type == 'logistic':
            y_pred = phi(y_pred)
            y_labels = np.ones(n)
            y_labels[y_pred < 0.5] = -1
            TP = sum(y[y_labels == 1] == 1)
            FP = sum(y[y_labels == -1] == 1)
            TN = sum(y[y_labels == -1] == -1)
            FN = sum(y[y_labels == 1] == -1)
            PC = (TP+TN)/float(TP+FP+FN+TN)
        elif model.type == 'mlogistic':
            y_pred = phi(y_pred.ravel())
            y_pred = y_pred.reshape(n, m)
            y_labels = y_pred.argmax(axis=1)+1
            PC = np.zeros(m)
            cls_set = np.arange(1, m+1)
            for ii in range(0, m):
                cls_sub = np.delete(cls_set, ii)
                TP = sum(y[y_labels == (ii+1)] == (ii+1))
                FP = sum(y[np.in1d(y_labels, cls_sub)] == (ii+1))
                TN = sum(y[np.in1d(y_labels, cls_sub)] ==
                         y_labels[np.in1d(y_labels, cls_sub)])
                FN = sum(np.in1d(y[y_labels == (ii+1)], cls_sub))
                PC[ii] = (TP+TN)/float(TP+FP+FN+TN)

            PC = sum(y == y_labels) / float(y_labels.size)

    if model.type == 'linear':
        prediction = collections.namedtuple('prediction', ['y_pred', 'SSE'])
        out = prediction(y_pred, SSE)
    elif model.type == 'logistic':
        prediction = collections.namedtuple('prediction', ['y_prob',
                                            'y_labels', 'PC'])
        out = prediction(y_pred, y_labels, PC)
    elif model.type == 'mlogistic':
        prediction = collections.namedtuple('prediction', ['y_prob',
                                            'y_labels', 'PC'])
        out = prediction(y_pred, y_labels, PC)

    return out


# helper functions for linear regression
def regression_warp(beta, time, q, y, alpha):
    """
    calculates optimal warping for function linear regression

    :param beta: numpy ndarray of shape (M,N) of M functions with N samples
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples
    responses
    :param alpha: numpy scalar

    :rtype: numpy array
    :return gamma_new: warping function

    """
    gam_M = uf.optimum_reparam(beta, time, q)
    qM = uf.warp_q_gamma(time, q, gam_M)
    y_M = trapz(qM * beta, time)

    gam_m = uf.optimum_reparam(-1 * beta, time, q)
    qm = uf.warp_q_gamma(time, q, gam_m)
    y_m = trapz(qm * beta, time)

    if y > alpha + y_M:
        gamma_new = gam_M
    elif y < alpha + y_m:
        gamma_new = gam_m
    else:
        gamma_new = uf.zero_crossing(y - alpha, q, beta, time, y_M, y_m,
                                     gam_M, gam_m)

    return gamma_new


# helper functions for logistic regression
def logistic_warp(beta, time, q, y):
    """
    calculates optimal warping for function logistic regression

    :param beta: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return gamma: warping function

    """
    if y == 1:
        gamma = uf.optimum_reparam(beta, time, q)
    elif y == -1:
        gamma = uf.optimum_reparam(-1*beta, time, q)
    return gamma


def phi(t):
    """
    calculates logistic function, returns 1 / (1 + exp(-t))

    :param t: scalar

    :rtype: numpy array
    :return out: return value

    """
    # logistic function, returns 1 / (1 + exp(-t))
    idx = t > 0
    out = np.empty(t.size, dtype=np.float)
    out[idx] = 1. / (1 + np.exp(-t[idx]))
    exp_t = np.exp(t[~idx])
    out[~idx] = exp_t / (1. + exp_t)
    return out


def logit_loss(b, X, y):
    """
    logistic loss function, returns Sum{-log(phi(t))}

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) of N responses

    :rtype: numpy array
    :return out: loss value

    """
    z = X.dot(b)
    yz = y * z
    idx = yz > 0
    out = np.zeros_like(yz)
    out[idx] = np.log(1 + np.exp(-yz[idx]))
    out[~idx] = (-yz[~idx] + np.log(1 + np.exp(yz[~idx])))
    out = out.sum()
    return out


def logit_gradient(b, X, y):
    """
    calculates gradient of the logistic loss

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return grad: gradient of logistic loss

    """
    z = X.dot(b)
    z = phi(y * z)
    z0 = (z - 1) * y
    grad = X.T.dot(z0)
    return grad


def logit_hessian(s, b, X, y):
    """
    calculates hessian of the logistic loss

    :param s: numpy ndarray of shape (M,N) of N functions with M samples
    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return out: hessian of logistic loss

    """
    z = X.dot(b)
    z = phi(y * z)
    d = z * (1 - z)
    wa = d * X.dot(s)
    Hs = X.T.dot(wa)
    out = Hs
    return out


# helper functions for multinomial logistic regression
def mlogit_warp_grad(alpha, beta, time, q, y, max_itr=8000, tol=1e-10,
                     delta=0.008, display=0):
    """
    calculates optimal warping for functional multinomial logistic regression

    :param alpha: scalar
    :param beta: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size M describing the sample points
    :param q: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses
    :param max_itr: maximum number of iterations (Default=8000)
    :param tol: stopping tolerance (Default=1e-10)
    :param delta: gradient step size (Default=0.008)
    :param display: display iterations (Default=0)

    :rtype: tuple of numpy array
    :return gam_old: warping function

    """

    gam_old = mw.mlogit_warp(np.ascontiguousarray(alpha),
                             np.ascontiguousarray(beta),
                             time, np.ascontiguousarray(q),
                             np.ascontiguousarray(y, dtype=np.int32), max_itr,
                             tol, delta, display)

    return gam_old


def mlogit_loss(b, X, Y):
    """
    calculates multinomial logistic loss (negative log-likelihood)

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return nll: negative log-likelihood

    """
    N, m = Y.shape  # n_samples, n_classes
    M = X.shape[1]  # n_features
    B = b.reshape(M, m)
    Yhat = np.dot(X, B)
    Yhat -= Yhat.min(axis=1)[:, np.newaxis]
    Yhat = np.exp(-Yhat)
    # l1-normalize
    Yhat /= Yhat.sum(axis=1)[:, np.newaxis]

    Yhat = Yhat * Y
    nll = np.sum(np.log(Yhat.sum(axis=1)))
    nll /= -float(N)

    return nll


def mlogit_gradient(b, X, Y):
    """
    calculates gradient of the multinomial logistic loss

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return grad: gradient

    """
    N, m = Y.shape  # n_samples, n_classes
    M = X.shape[1]  # n_features
    B = b.reshape(M, m)
    Yhat = np.dot(X, B)
    Yhat -= Yhat.min(axis=1)[:, np.newaxis]
    Yhat = np.exp(-Yhat)
    # l1-normalize
    Yhat /= Yhat.sum(axis=1)[:, np.newaxis]

    _Yhat = Yhat * Y
    _Yhat /= _Yhat.sum(axis=1)[:, np.newaxis]
    Yhat -= _Yhat
    grad = np.dot(X.T, Yhat)
    grad /= -float(N)
    grad = grad.ravel()

    return grad



"""
geometry functions for SRSF Manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

from numpy import arccos, sin, cos, linspace, zeros, sqrt
from scipy.integrate import trapz

def inv_exp_map(Psi, psi):
    tmp = inner_product(Psi,psi)
    if tmp > 1:
        tmp = 1
    if tmp < -1:
        tmp = -1

    theta = arccos(tmp)

    if (theta < 1e-10):
        exp_inv = zeros(psi.shape[0])
    else: 
        exp_inv = theta / sin(theta) * (psi - cos(theta)*Psi)
    
    return exp_inv, theta


def exp_map(psi, v):
    v_norm = L2norm(v)
    expgam = cos(v_norm) * psi + sin(v_norm) * v / v_norm
    return expgam

def inner_product(psi1, psi2):
    M = psi1.shape[0]
    t = linspace(0,1,M)
    ip = trapz(psi1*psi2, t)
    return ip

def L2norm(psi):
    M = psi.shape[0]
    t = linspace(0,1,M)
    l2norm = sqrt(trapz(psi*psi,t))
    return l2norm




from libcpp cimport bool
cdef extern from "gropt/incl/ElasticCurvesReparam.h":
    void optimum_reparam(double *C1, double *C2, int n, int d, double w,
                     bool onlyDP, bool rotated, bool isclosed, int skipm,
                     int autoselectC, double *opt, bool swap, double *fopts,
                     double *comtime);



import numpy as np
cimport numpy as np
cimport cython

from cython.operator cimport dereference as deref

from libcpp cimport bool

cdef extern from "armadillo" namespace "arma" nogil:
    # matrix class (double)
    cdef cppclass mat:
        mat(double * aux_mem, int n_rows, int n_cols, bool copy_aux_mem, bool strict) nogil
        mat(double * aux_mem, int n_rows, int n_cols) nogil
        mat(int n_rows, int n_cols) nogil
        mat() nogil
        # attributes
        int n_rows
        int n_cols
        int n_elem
        int n_slices
        int n_nonzero
        # fuctions
        mat i() nogil #inverse
        mat t() nogil #transpose
        vec diag() nogil
        vec diag(int) nogil
        fill(double) nogil
        void raw_print(char*) nogil
        void raw_print() nogil
        vec unsafe_col(int) nogil
        vec col(int) nogil
        #print(char)
        #management
        mat reshape(int, int) nogil
        mat resize(int, int) nogil
        double * memptr() nogil
        # opperators
        double& operator[](int) nogil
        double& operator[](int,int) nogil
        double& at(int,int) nogil
        double& at(int) nogil
        mat operator*(mat) nogil
        mat operator%(mat) nogil
        vec operator*(vec) nogil
        mat operator+(mat) nogil
        mat operator-(mat) nogil
        mat operator*(double) nogil
        mat operator-(double) nogil
        mat operator+(double) nogil
        mat operator/(double) nogil
        #etc

    cdef cppclass cube:
        cube(double * aux_mem, int n_rows, int n_cols, int n_slices, bool copy_aux_mem, bool strict) nogil
        cube(double * aux_mem, int n_rows, int n_cols, int n_slices) nogil
        cube(int, int, int) nogil
        cube() nogil
        #attributes
        int n_rows
        int n_cols
        int n_elem
        int n_elem_slices
        int n_slices
        int n_nonzero
        double * memptr() nogil
        void raw_print(char*) nogil
        void raw_print() nogil
        

    # vector class (double)
    cdef cppclass vec:
        cppclass iterator:
            double& operator*()
            iterator operator++()
            iterator operator--()
            iterator operator+(size_t)
            iterator operator-(size_t)
            bint operator==(iterator)
            bint operator!=(iterator)
            bint operator<(iterator)
            bint operator>(iterator)
            bint operator<=(iterator)
            bint operator>=(iterator)
        cppclass reverse_iterator:
            double& operator*()
            iterator operator++()
            iterator operator--()
            iterator operator+(size_t)
            iterator operator-(size_t)
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
            bint operator<(reverse_iterator)
            bint operator>(reverse_iterator)
            bint operator<=(reverse_iterator)
            bint operator>=(reverse_iterator)
        vec(double * aux_mem, int number_of_elements, bool copy_aux_mem, bool strict) nogil
        vec(double * aux_mem, int number_of_elements) nogil
        vec(int) nogil
        vec() nogil
        # attributes
        int n_elem
        # opperators
        double& operator[](int)
        double& at(int)
        vec operator%(vec)
        vec operator+(vec)
        vec operator/(vec)
        vec operator*(mat)
        vec operator*(double)
        vec operator-(double)
        vec operator+(double)
        vec operator/(double)
        iterator begin()
        iterator end()
        reverse_iterator rbegin()
        reverse_iterator rend()


        # functions
        double * memptr()
        void raw_print(char*) nogil
        void raw_print() nogil

    # Armadillo Linear Algebra tools
    cdef bool chol(mat R, mat X) nogil # preallocated result
    cdef mat chol(mat X) nogil # new result
    cdef bool inv(mat R, mat X) nogil
    cdef mat inv(mat X) nogil
    cdef bool solve(vec x, mat A, vec b) nogil
    cdef vec solve(mat A, vec b) nogil
    cdef bool solve(mat X, mat A, mat B) nogil
    cdef mat solve(mat A, mat B) nogil
    cdef bool eig_sym(vec eigval, mat eigvec, mat B) nogil
    cdef bool svd(mat U, vec s, mat V, mat X, method) nogil
    cdef bool lu(mat L, mat U, mat P, mat X) nogil
    cdef bool lu(mat L, mat U, mat X) nogil
    cdef mat pinv(mat A) nogil
    cdef bool pinv(mat B, mat A) nogil
    cdef bool qr(mat Q, mat R, mat X) nogil
    cdef float dot(vec a, vec b) nogil
    cdef mat arma_cov "cov"(mat X) nogil
    cdef vec arma_mean "mean"(mat X, int dim) nogil
    cdef mat arma_var "var"(mat X, int norm_type, int dim) nogil

## TODO: refactor to return pointer, but other function derefed. 

##### Tools to convert numpy arrays to armadillo arrays ######
cdef mat * numpy_to_mat(np.ndarray[np.double_t, ndim=2] X):
    if not (X.flags.f_contiguous or X.flags.owndata):
        X = X.copy(order="F")
    cdef mat *aR_p  = new mat(<double*> X.data, X.shape[0], X.shape[1], False, True)
    return aR_p

cdef mat numpy_to_mat_d(np.ndarray[np.double_t, ndim=2] X):
    cdef mat * aR_p = numpy_to_mat(X)
    cdef mat aR = deref(aR_p)
    del aR_p
    return aR

cdef cube * numpy_to_cube(np.ndarray[np.double_t, ndim=3] X):
    cdef cube *aR_p
    if not X.flags.c_contiguous:
        raise ValueError("For Cube, numpy array must be C contiguous")
    aR_p  = new cube(<double*> X.data, X.shape[2], X.shape[1], X.shape[0], False, True)
    return aR_p

cdef cube numpy_to_cube_d(np.ndarray[np.double_t, ndim=3] X):
    cdef cube * aR_p = numpy_to_cube(X)
    cdef cube aR = deref(aR_p)
    del aR_p
    return aR
    
cdef vec * numpy_to_vec(np.ndarray[np.double_t, ndim=1] x):
    if not (x.flags.f_contiguous or x.flags.owndata):
        x = x.copy()
    cdef vec *ar_p = new vec(<double*> x.data, x.shape[0], False, True)
    return ar_p

cdef vec numpy_to_vec_d(np.ndarray[np.double_t, ndim=1] x):
    cdef vec *ar_p = numpy_to_vec(x)
    cdef vec ar = deref(ar_p)
    del ar_p
    return ar
    

#### Get subviews #####
cdef vec * mat_col_view(mat * x, int col) nogil:
    cdef vec * ar_p = new vec(x.memptr()+x.n_rows*col, x.n_rows, False, True)
    return ar_p

cdef vec mat_col_view_d(mat * x, int col) nogil:
    cdef vec * ar_p = mat_col_view(x, col)
    cdef vec ar = deref(ar_p)
    del ar_p
    return ar

cdef mat * cube_slice_view(cube * x, int slice) nogil:
    cdef mat *ar_p = new mat(x.memptr() + x.n_rows*x.n_cols*slice,
                           x.n_rows, x.n_cols, False, True)
    return ar_p

cdef mat cube_slice_view_d(cube * x, int slice) nogil:
    cdef mat * ar_p = cube_slice_view(x, slice)
    cdef mat ar = deref(ar_p)
    del ar_p
    return ar



##### Converting back to python arrays, must pass preallocated memory or None
# all data will be copied since numpy doesn't own the data and can't clean up
# otherwise. Maybe this can be improved. #######
@cython.boundscheck(False)
cdef np.ndarray[np.double_t, ndim=2] mat_to_numpy(const mat & X, np.ndarray[np.double_t, ndim=2] D):
    cdef const double * Xptr = X.memptr()
    
    if D is None:
        D = np.empty((X.n_rows, X.n_cols), dtype=np.double, order="F")
    cdef double * Dptr = <double*> D.data
    for i in range(X.n_rows*X.n_cols):
        Dptr[i] = Xptr[i]
    return D

@cython.boundscheck(False)
cdef np.ndarray[np.double_t, ndim=1] vec_to_numpy(const vec & X, np.ndarray[np.double_t, ndim=1] D):
    cdef const double * Xptr = X.memptr()
    
    if D is None:
        D = np.empty(X.n_elem, dtype=np.double)
    cdef double * Dptr = <double*> D.data
    for i in range(X.n_elem):
        Dptr[i] = Xptr[i]
    return D

### A few wrappers for much needed numpy linalg functionality using armadillo
# cpdef np_chol(np.ndarray[np.double_t, ndim=2] X):
#     # initialize result numpy array
#     cdef np.ndarray[np.double_t, ndim=2] R = \
#          np.empty((X.shape[0], X.shape[1]), dtype=np.double, order="F")
#     # wrap them up in armidillo arrays
#     cdef mat *aX = new mat(<double*> X.data, X.shape[0], X.shape[1], False, True)
#     cdef mat *aR  = new mat(<double*> R.data, R.shape[0], R.shape[1], False, True)
    
#     chol(deref(aR), deref(aX))
    
#     return R

# cpdef np_inv(np.ndarray[np.double_t, ndim=2] X):
#     # initialize result numpy array
#     cdef np.ndarray[np.double_t, ndim=2] R = \
#          np.empty((X.shape[0], X.shape[1]), dtype=np.double, order="F")
#     # wrap them up in armidillo arrays
#     cdef mat *aX = new mat(<double*> X.data, X.shape[0], X.shape[1], False, True)
#     cdef mat *aR  = new mat(<double*> R.data, R.shape[0], R.shape[1], False, True)
    
#     inv(deref(aR), deref(aX))
    
#     return R


# def np_eig_sym(np.ndarray[np.double_t, ndim=2] X):
#     # initialize result numpy array
#     cdef np.ndarray[np.double_t, ndim=2] R = \
#          np.empty((X.shape[0], X.shape[1]), dtype=np.double, order="F")
#     cdef np.ndarray[np.double_t, ndim=1] v = \
#          np.empty(X.shape[0], dtype=np.double)
#     # wrap them up in armidillo arrays
#     cdef mat *aX = new mat(<double*> X.data, X.shape[0], X.shape[1], False, True)
#     cdef mat *aR  = new mat(<double*> R.data, R.shape[0], R.shape[1], False, True)
#     cdef vec *av = new vec(<double*> v.data, v.shape[0], False, True)

#     eig_sym(deref(av), deref(aR), deref(aX))

#     return [v, R]


# def example(np.ndarray[np.double_t, ndim=2] X):
#     cdef mat aX = numpy_to_mat(X)
#     cdef mat XX = aX.t() * aX
#     cdef mat ch = chol(XX)
#     ch.raw_print()
#     cdef np.ndarray[np.double_t,ndim=2] Y = mat_to_numpy(ch, None)
#     return Y



cimport cfPLS
import numpy as np

cimport numpy as np
from cpython cimport array

def fpls_warp(np.ndarray[double, ndim=1, mode="c"] time, np.ndarray[double, ndim=2, mode="c"] gam,
              np.ndarray[double, ndim=2, mode="c"] qf, np.ndarray[double, ndim=2, mode="c"] qg,
              np.ndarray[double, ndim=1, mode="c"] wf, np.ndarray[double, ndim=1, mode="c"] wg,
              max_iter = 100, tol = 1e-4, delta = 0.1, display = 1):
    """
    cython interface perform warping calculation for PLS cost function

    :param time: vector of size N describing the sample points
    :param gam: numpy ndarray of shape (M,N) of N init warping functions with M samples
    :param qf: numpy ndarray of shape (M,N) of N srsfs with M samples
    :param qg: numpy ndarray of shape (M,N) of N srsfs with M samples
    :param wf: numpy ndarray of shape (M,1) weight function f
    :param wg: numpy ndarray of shape (M,1) weight function f
    :param max_iter: maximal number of iterations (default = 100)
    :param tol: stopping tolerance (default = 1e-4)
    :param delta: step size (default = .1)
    :param display: show iterations (default = 1)

    :rtype numpy ndarray
    :return gamo: describing the warping functions

    """
    cdef int m1, n1, max_itri, displayi
    cdef double toli, deltai
    toli = tol
    deltai = delta
    max_itri = max_iter
    displayi = display
    m1 = gam.shape[0]
    n1 = gam.shape[1]
    cdef np.ndarray[double, ndim=1, mode="c"] gam1 = np.zeros(m1 * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] qf1 = np.zeros(m1 * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] qg1 = np.zeros(m1 * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] gamout = np.zeros(m1 * n1)
    for ii in xrange(0, n1):
        gam1[ii * m1:ii * m1 + m1] = gam[:, ii]
        qf1[ii * m1:ii * m1 + m1] = qf[:, ii]
        qg1[ii * m1:ii * m1 + m1] = qg[:, ii]

    gam1 = np.ascontiguousarray(gam1)
    qf1 = np.ascontiguousarray(qf1)
    qg1 = np.ascontiguousarray(qg1)
    gamout = np.ascontiguousarray(gamout)

    cfPLS.fpls_warp_grad(&m1, &n1, &time[0], &gam1[0], &qf1[0], &qg1[0], &wf[0], &wg[0], &max_itri, &toli, &deltai,
                         &displayi, &gamout[0])

    gamo = np.zeros((m1, n1))
    for ii in xrange(0, n1):
        gamo[:, ii] = gamout[ii * m1:ii * m1 + m1]

    return gamo



import numpy as np
cimport cbayesian
cimport cyarma
cimport numpy as np

include "cyarma.pyx"

from libcpp cimport bool

def bcalcY(area, np.ndarray[double, ndim=1, mode="c"] y):
    y = np.ascontiguousarray(y)
    cdef vec ay = numpy_to_vec_d(y)
    cdef vec out = calcY(area, ay)

    cdef np.ndarray[np.double_t,ndim=1] out1 = vec_to_numpy(out, None)

    return(out1)

def bcuL2norm2(np.ndarray[double, ndim=1, mode="c"] x, np.ndarray[double, ndim=1, mode="c"] y):
    x = np.ascontiguousarray(x)
    y = np.ascontiguousarray(y)
    cdef vec ax = numpy_to_vec_d(x)
    cdef vec ay = numpy_to_vec_d(y)
    cdef vec out = cuL2norm2(ax, ay)

    cdef np.ndarray[np.double_t,ndim=1] out1 = vec_to_numpy(out, None)

    return(out1)

def ctrapzCpp(np.ndarray[double, ndim=1, mode="c"] x, np.ndarray[double, ndim=1, mode="c"] y):
    x = np.ascontiguousarray(x)
    y = np.ascontiguousarray(y)
    cdef vec ax = numpy_to_vec_d(x)
    cdef vec ay = numpy_to_vec_d(y)

    cdef double out = trapzCpp(ax, ay)

    return(out)

def border_l2norm(np.ndarray[double, ndim=1, mode="c"] x, np.ndarray[double, ndim=1, mode="c"] y):
    x = np.ascontiguousarray(x)
    y = np.ascontiguousarray(y)
    cdef vec ax = numpy_to_vec_d(x)
    cdef vec ay = numpy_to_vec_d(y)

    cdef double out = order_l2norm(ax, ay)

    return(out)



cimport cDP
import numpy as np
from numpy.linalg import norm

cimport numpy as np
from cpython cimport array

def coptimum_reparam_N(np.ndarray[double, ndim=1, mode="c"] mq, np.ndarray[double, ndim=1, mode="c"] time,
                      np.ndarray[double, ndim=2, mode="c"] q, lam1=0.0):
    """
    cython interface calculates the warping to align a set of SRSFS q to a single SRSF mq

    :param mq: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of N srsfs with M samples
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of q with mq

    """
    cdef int M, N, n1, disp
    cdef double lam
    mq = mq / norm(mq)
    M, N = q.shape[0], q.shape[1]
    n1 = 1
    disp = 0
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] qi = np.zeros(M)

    gam = np.zeros((M, N))
    for k in xrange(0, N):
        qi = q[:, k] / norm(q[:, k])
        qi = np.ascontiguousarray(qi)

        cDP.DP(&qi[0], &mq[0], &n1, &M, &lam, &disp, &gami[0])
        gam[:, k] = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam

def coptimum_reparam_N2(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                       np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface calculates the warping to align a set of SRSFs q1 to another set of SRSFs q2

    :param q1: numpy ndarray of shape (M,N) of M srsfs with N samples
    :param time: vector of size N describing the sample points
    :param q2: numpy ndarray of shape (M,N) of M srsfs with N samples
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of q with mq

    """
    cdef int M, N, n1, disp
    cdef double lam

    M, N = q1.shape[0], q1.shape[1]
    n1 = 1
    lam = lam1
    disp = 0
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M)

    gam = np.zeros((M, N))
    for k in xrange(0, N):
        q1i = q1[:, k] / norm(q1[:, k])
        q2i = q2[:, k] / norm(q2[:, k])
        q1i = np.ascontiguousarray(q1i)
        q2i = np.ascontiguousarray(q2i)

        cDP.DP(&q2i[0], &q1i[0], &n1, &M, &lam, &disp, &gami[0])
        gam[:, k] = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam

def coptimum_reparam(np.ndarray[double, ndim=1, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=1, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align SRSFs q2 to q1

    :param q1: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q2: vector of size N samples of second SRSF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, n1, disp
    cdef double lam
    M = q1.shape[0]
    n1 = 1
    lam = lam1
    disp = 0
    q1 = q1 / norm(q1)
    q2 = q2 / norm(q2)
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)

    cDP.DP(&q2[0], &q1[0], &n1, &M, &lam, &disp, &gami[0])
    gam = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam

def coptimum_reparam_N2_pair(np.ndarray[double, ndim=2, mode="c"] q, np.ndarray[double, ndim=1, mode="c"] time,
                            np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align paired SRSF f1 and f2 to q

    :param q: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q1: vector of size N samples of second SRSF
    :param q2: vector of size N samples of second SRSF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, N, n1, disp
    n1 = 2
    cdef double lam
    M, N = q1.shape[0], q1.shape[1]
    lam = lam1
    disp = 0
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M * n1)

    gam = np.zeros((M, N))
    for k in xrange(0, N):
        q1i = q.reshape(M*n1)
        q2tmp = np.column_stack((q1[:, k], q2[:, k]))
        q2i = q2tmp.reshape(M*n1)

        q1i = np.ascontiguousarray(q1i)
        q2i = np.ascontiguousarray(q2i)

        cDP.DP(&q2i[0], &q1i[0], &n1, &M, &lam, &disp, &gami[0])
        gam[:, k] = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam

def coptimum_reparam_pair_q(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                          np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align paired srsf q2 to q1

    :param q1: vector of size N samples of first function
    :param time: vector of size N describing the sample points
    :param q2: vector of size N samples of second function
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align f2 with f1
    """
    cdef int M, N, disp
    cdef double lam
    M, N = q1.shape[0], q1.shape[1]
    lam = lam1
    disp = 0
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M * N)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M * N)

    sizes = np.zeros(1, dtype=np.int32)
    q1i = q1.reshape(M*N)
    q2i = q2.reshape(M*N)

    q1i = np.ascontiguousarray(q1i)
    q2i = np.ascontiguousarray(q2i)

    cDP.DP(&q2i[0], &q1i[0], &N, &M, &lam, &disp, &gami[0])
    gam = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam

def coptimum_reparam_curve(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align curve q2 to q1

    :param q1: matrix of size nxN samples of first SRVF
    :param time: vector of size N describing the sample points
    :param q2: matrix of size nxN samples of second SRVF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align f2 with f1
    """
    cdef int M, n1, disp
    cdef double lam
    n1 = q1.shape[0]
    M = q1.shape[1]
    lam = lam1
    disp = 0
    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M * n1)

    q1i = q1.reshape(M*n1, order='F')
    q2i = q2.reshape(M*n1, order='F')

    q1i = np.ascontiguousarray(q1i)
    q2i = np.ascontiguousarray(q2i)

    cDP.DP(&q2i[0], &q1i[0], &n1, &M, &lam, &disp, &gami[0])
    gam = (gami - gami[0]) / (gami[-1] - gami[0])

    return gam



cimport cDPQ
import numpy as np
from numpy.linalg import norm

cimport numpy as np
from cpython cimport array

def coptimum_reparamN(np.ndarray[double, ndim=1, mode="c"] mq, np.ndarray[double, ndim=1, mode="c"] time,
                      np.ndarray[double, ndim=2, mode="c"] q, lam1=0.0):
    """
    cython interface calculates the warping to align a set of srfs q to a single srsf mq

    :param mq: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of N srsfs with M samples
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of q with mq

    """
    cdef int M, N, n1
    cdef double lam
    mq = mq / norm(mq)
    M, N = q.shape[0], q.shape[1]
    n1 = 1
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] qi = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    gam = np.zeros((M, N))
    sizes = np.zeros(N, dtype=np.int32)
    Go = np.zeros((M, N))
    To = np.zeros((M, N))
    for k in xrange(0, N):
        qi = q[:, k] / norm(q[:, k])
        qi = np.ascontiguousarray(qi)

        cDPQ.DynamicProgrammingQ2(&mq[0], &time[0], &qi[0], &time[0], n1, M, M, &time[0], &time[0], M, M, &G[0],
                                  &T[0], &size[0], lam)
        sizes[k] = np.int32(size)
        Go[:, k] = G
        To[:, k] = T

    for k in xrange(0, N):
        gam0 = np.interp(time, To[0:sizes[k], k], Go[0:sizes[k], k])
        gam[:, k] = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparamN2(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                       np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface calculates the warping to align a set of srsfs q1 to another set of srsfs q2

    :param q1: numpy ndarray of shape (M,N) of M srsfs with N samples
    :param time: vector of size N describing the sample points
    :param q2: numpy ndarray of shape (M,N) of M srsfs with N samples
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype numpy ndarray
    :return gam: describing the warping functions used to align columns of q with mq

    """
    cdef int M, N, n1
    cdef double lam

    M, N = q1.shape[0], q1.shape[1]
    n1 = 1
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    gam = np.zeros((M, N))
    sizes = np.zeros(N, dtype=np.int32)
    Go = np.zeros((M, N))
    To = np.zeros((M, N))
    for k in xrange(0, N):
        q1i = q1[:, k] / norm(q1[:, k])
        q2i = q2[:, k] / norm(q2[:, k])
        q1i = np.ascontiguousarray(q1i)
        q2i = np.ascontiguousarray(q2i)

        cDPQ.DynamicProgrammingQ2(&q1i[0], &time[0], &q2i[0], &time[0], n1, M, M, &time[0], &time[0], M, M, &G[0],
                                  &T[0], &size[0], lam)
        sizes[k] = np.int32(size)
        Go[:, k] = G
        To[:, k] = T

    for k in xrange(0, N):
        gam0 = np.interp(time, To[0:sizes[k], k], Go[0:sizes[k], k])
        gam[:, k] = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparam(np.ndarray[double, ndim=1, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=1, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align srsf q2 to q1

    :param q1: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q2: vector of size N samples of second SRSF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, n1
    cdef double lam
    M = q1.shape[0]
    n1 = 1
    lam = lam1
    q1 = q1 / norm(q1)
    q2 = q2 / norm(q2)
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    sizes = np.zeros(1, dtype=np.int32)
    Go = np.zeros((M, 1))
    To = np.zeros((M, 1))
    cDPQ.DynamicProgrammingQ2(&q1[0], &time[0], &q2[0], &time[0], n1, M, M, &time[0], &time[0], M, M, &G[0],
                              &T[0], &size[0], lam)
    sizes = np.int32(size)
    Go[:, 0] = G
    To[:, 0] = T
    gam0 = np.interp(time, To[0:sizes[0], 0], Go[0:sizes[0], 0])
    gam = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparamN2_pair(np.ndarray[double, ndim=2, mode="c"] q, np.ndarray[double, ndim=1, mode="c"] time,
                            np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align paired srsf q1 and q2 to q

    :param q: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q1: vector of size N samples of second SRSF
    :param q2: vector of size N samples of second SRSF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, N, n1
    n1 = 2
    cdef double lam
    M, N = q1.shape[0], q1.shape[1]
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    gam = np.zeros((M, N))
    sizes = np.zeros(N, dtype=np.int32)
    Go = np.zeros((M, N))
    To = np.zeros((M, N))
    for k in xrange(0, N):
        q1i = q.reshape(M*n1)
        q2tmp = np.column_stack((q1[:, k], q2[:, k]))
        q2i = q2tmp.reshape(M*n1)

        q1i = np.ascontiguousarray(q1i)
        q2i = np.ascontiguousarray(q2i)

        cDPQ.DynamicProgrammingQ2(&q1i[0], &time[0], &q2i[0], &time[0], n1, M, M, &time[0], &time[0], M, M, &G[0],
                                  &T[0], &size[0], lam)
        sizes[k] = np.int32(size)
        Go[:, k] = G
        To[:, k] = T

    for k in xrange(0, N):
        gam0 = np.interp(time, To[0:sizes[k], k], Go[0:sizes[k], k])
        gam[:, k] = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparam_pair(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                          np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align paired srsf q2 to q1

    :param q1: vector of size N samples of first SRSF
    :param time: vector of size N describing the sample points
    :param q2: vector of size N samples of second SRSF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, N
    cdef double lam
    M, N = q1.shape[0], q1.shape[1]
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M * N)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M * N)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    sizes = np.zeros(1, dtype=np.int32)
    q1i = q1.reshape(M*N)
    q2i = q2.reshape(M*N)

    q1i = np.ascontiguousarray(q1i)
    q2i = np.ascontiguousarray(q2i)

    Go = np.zeros((M, 1))
    To = np.zeros((M, 1))
    cDPQ.DynamicProgrammingQ2(&q1i[0], &time[0], &q2i[0], &time[0], N, M, M, &time[0], &time[0], M, M, &G[0],
                              &T[0], &size[0], lam)
    sizes = np.int32(size)
    Go[:, 0] = G
    To[:, 0] = T
    gam0 = np.interp(time, To[0:sizes[0], 0], Go[0:sizes[0], 0])
    gam = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam

def coptimum_reparam_curve(np.ndarray[double, ndim=2, mode="c"] q1, np.ndarray[double, ndim=1, mode="c"] time,
                     np.ndarray[double, ndim=2, mode="c"] q2, lam1=0.0):
    """
    cython interface for calculates the warping to align srvf q2 to q1

    :param q1: matrix of size nxN samples of first SRVF
    :param time: vector of size N describing the sample points
    :param q2: matrix of size nxN samples of second SRVF
    :param lam1: controls the amount of elasticity (default = 0.0)

    :rtype vector
    :return gam: describing the warping function used to align q2 with q1
    """
    cdef int M, n1
    cdef double lam
    n1 = q1.shape[0]
    M = q1.shape[1]
    lam = lam1
    cdef np.ndarray[double, ndim=1, mode="c"] G = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] T = np.zeros(M)
    cdef np.ndarray[double, ndim=1, mode="c"] q1i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] q2i = np.zeros(M * n1)
    cdef np.ndarray[double, ndim=1, mode="c"] size = np.zeros(1)

    q1i = q1.reshape(M*n1, order='F')
    q2i = q2.reshape(M*n1, order='F')

    q1i = np.ascontiguousarray(q1i)
    q2i = np.ascontiguousarray(q2i)

    sizes = np.zeros(1, dtype=np.int32)
    Go = np.zeros((M, 1))
    To = np.zeros((M, 1))
    cDPQ.DynamicProgrammingQ2(&q1i[0], &time[0], &q2i[0], &time[0], n1, M, M, &time[0], &time[0], M, M, &G[0],
                              &T[0], &size[0], lam)
    sizes = np.int32(size)
    Go[:, 0] = G
    To[:, 0] = T
    gam0 = np.interp(time, To[0:sizes[0], 0], Go[0:sizes[0], 0])
    gam = (gam0 - gam0[0]) / (gam0[-1] - gam0[0])

    return gam



"""
functions for SRVF curve manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

from scipy.interpolate import InterpolatedUnivariateSpline, interp1d
from scipy.integrate import trapz, cumtrapz
from numpy import zeros, ones, cumsum, linspace, gradient, sqrt, ascontiguousarray
from numpy import finfo, double, eye, roll, tile, vstack, array, cos, sin
from numpy import arccos, fabs
from scipy.linalg import norm, svd, det, solve
import optimum_reparam_N as orN
import fdasrsf.utility_functions as uf


def resamplecurve(x, N=100, mode='O'):
    """
    This function resamples a curve to have N samples

    :param x: numpy ndarray of shape (2,M) of M samples
    :param N: Number of samples for new curve (default = 100)
    :param mode: Open ('O') or closed curve ('C') (default 'O')

    :rtype: numpy ndarray
    :return xn: resampled curve

    """
    n, T = x.shape
    xn = zeros((n, N))

    delta = zeros(T)
    for r in range(1, T):
        delta[r] = norm(x[:, r] - x[:, r - 1])

    cumdel = cumsum(delta) / delta.sum()
    newdel = linspace(0, 1, N)

    for r in range(0, n):
        s = InterpolatedUnivariateSpline(cumdel, x[r, :], k=3)
        xn[r, :] = s(newdel)

    if mode == 'C':
        q = curve_to_q(xn)
        qn = project_curve(q)
        xn = q_to_curve(qn)

    return (xn)


def calculatecentroid(beta):
    """
    This function calculates centroid of a parameterized curve

    :param beta: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return centroid: center coordinates

    """
    n, T = beta.shape
    betadot = gradient(beta, 1. / (T - 1))
    betadot = betadot[1]
    normbetadot = zeros(T)
    integrand = zeros((n, T))
    for i in range(0, T):
        normbetadot[i] = norm(betadot[:, i])
        integrand[:, i] = beta[:, i] * normbetadot[i]

    scale = trapz(normbetadot, linspace(0, 1, T))
    centroid = trapz(integrand, linspace(0, 1, T), axis=1) / scale

    return (centroid)


def curve_to_q(beta,scale=True,mode='O'):
    """
    This function converts curve beta to srvf q

    :param beta: numpy ndarray of shape (2,M) of M samples
    :param scale: scale curve to length 1
    :param mode: Open ('O') or closed curve ('C') (default 'O')

    :rtype: numpy ndarray
    :return q: srvf of curve
    :return len: length of curve

    """
    n, T = beta.shape
    v = gradient(beta, 1. / (T - 1))
    v = v[1]

    q = zeros((n, T))
    for i in range(0, T):
        L = sqrt(norm(v[:, i]))
        if L > 0.0001:
            q[:, i] = v[:, i] / L
        else:
            q[:, i] = v[:, i] * 0.0001

    len1 = sqrt(innerprod_q2(q,q))
    if scale:
        q = q / sqrt(innerprod_q2(q, q))
    
    if mode == 'C':
        q = project_curve(q)

    return q


def q_to_curve(q,scale=1):
    """
    This function converts srvf to beta

    :param q: numpy ndarray of shape (n,M) of M samples
    :param scale: scale of curve

    :rtype: numpy ndarray
    :return beta: parameterized curve

    """
    n,T = q.shape
    qnorm = zeros(T)
    for i in range(0, T):
        qnorm[i] = norm(q[:, i])

    beta = zeros((n,T))
    for i in range(0,n):
        beta[i,:] = cumtrapz(q[i, :] * qnorm * scale, initial=0)/T

    return (beta)


def Basis_Normal_A(q):
    """
    Find Normal Basis

    :param q: numpy ndarray (n,T) defining T points on n dimensional SRVF

    :rtype list
    :return delg: basis
    """
    n,T = q.shape
    e = eye(n)
    Ev = zeros((n,T,n))
    for i in range(0,n):
        Ev[:,:,i] = tile(e[:,i], (T, 1)).T
    
    qnorm = zeros(T)
    for t in range(0,T):
        qnorm[t] = norm(q[:,t])
    
    delG = list()
    for i in range(0,n):
        tmp1 = tile(q[i,:]/qnorm, (n,1))
        tmp2 = tile(qnorm, (n,1))
        delG.append(tmp1*q + tmp2*Ev[:,:,i])
    
    return delG


def optimum_reparam_curve(q1, q2, lam=0.0):
    """
    calculates the warping to align srsf q2 to q1

    :param q1: matrix of size nxN or array of NxM samples of first SRVF
    :param time: vector of size N describing the sample points
    :param q2: matrix of size nxN or array of NxM samples samples of second SRVF
    :param lam: controls the amount of elasticity (default = 0.0)

    :rtype: vector
    :return gam: describing the warping function used to align q2 with q1

    """
    time = linspace(0, 1, q1.shape[1])
    gam = orN.coptimum_reparam_curve(ascontiguousarray(q1), time,
                                     ascontiguousarray(q2), lam)

    return gam


def innerprod_q2(q1, q2):
    """
    This function calculates the inner product in srvf space

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return val: inner product

    """
    T = q1.shape[1]
    val = sum(sum(q1 * q2)) / T

    return (val)


def find_best_rotation(q1, q2):
    """
    This function calculates the best rotation between two srvfs using
    procustes rigid alignment

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return q2new: optimal rotated q2 to q1
    :return R: rotation matrix

    """
    eps = finfo(double).eps
    n, T = q1.shape
    A = q1.dot(q2.T)
    U, s, V = svd(A)

    if (abs(det(U) * det(V) - 1) < 10 * eps):
        S = eye(n)
    else:
        S = eye(n)
        S[:, -1] = -S[:, -1]

    R = U.dot(S).dot(V.T)
    q2new = R.dot(q2)

    return (q2new, R)


def calculate_variance(beta):
    """
    This function calculates variance of curve beta

    :param beta: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return variance: variance

    """
    n, T = beta.shape
    betadot = gradient(beta, 1. / (T - 1))
    betadot = betadot[1]
    normbetadot = zeros(T)
    centroid = calculatecentroid(beta)
    integrand = zeros((n, n, T))
    t = linspace(0, 1, T)
    for i in range(0, T):
        normbetadot[i] = norm(betadot[:, i])
        a1 = (beta[:, i] - centroid)
        a1 = a1.reshape((n, 1))
        integrand[:, :, i] = a1.dot(a1.T) * normbetadot[i]

    l = trapz(normbetadot, t)
    variance = trapz(integrand, t, axis=2)
    variance /= l

    return (variance)


def psi(x, a, q):
    """
    This function formats variance output

    :param x: numpy ndarray of shape (2,M) of M samples curve
    :param a: numpy ndarray of shape (2,1) mean
    :param q: numpy ndarray of shape (2,M) of M samples srvf

    :rtype: numpy ndarray
    :return psi1: variance
    :return psi2: cross variance
    :return psi3: curve end
    :return psi4: curve end

    """
    T = q.shape[1]
    covmat = calculate_variance(x + tile(a, [T, 1]).T)
    psi1 = covmat[0, 0] - covmat[1, 1]
    psi2 = covmat[0, 1]
    psi3 = x[0, -1]
    psi4 = x[1, -1]

    return (psi1, psi2, psi3, psi4)


def find_basis_normal(q):
    """
    Finds the basis normal to the srvf

    :param q1: numpy ndarray of shape (2,M) of M samples

    :rtype: list of numpy ndarray
    :return basis: list containing basis vectors

    """
    n, T = q.shape

    f1 = zeros((n, T))
    f2 = zeros((n, T))
    for i in range(0, T):
        f1[:, i] = q[0, i] * q[:, i] / norm(q[:, i]) + array([norm(q[:, i]),
                                                              0])
        f2[:, i] = q[1, i] * q[:, i] / norm(q[:, i]) + array([0,
                                                              norm(q[:, i])])

    h3 = f1
    h4 = f2
    integrandb3 = zeros(T)
    integrandb4 = zeros(T)
    for i in range(0, T):
        a = q[:, i].T
        integrandb3[i] = a.dot(h3[:, i])
        integrandb4[i] = a.dot(h4[:, i])

    b3 = h3 - q * trapz(integrandb3, linspace(0, 1, T))
    b4 = h4 - q * trapz(integrandb4, linspace(0, 1, T))

    basis = [b3, b4]

    return (basis)


def calc_j(basis):
    """
    Calculates Jacobian matrix from normal basis

    :param basis: list of numpy ndarray of shape (2,M) of M samples basis

    :rtype: numpy ndarray
    :return j: Jacobian

    """
    b1 = basis[0]
    b2 = basis[1]
    T = b1.shape[1]
    integrand11 = zeros(T)
    integrand12 = zeros(T)
    integrand22 = zeros(T)

    for i in range(0, T):
        a = b1[:, i].T
        b = b2[:, i].T
        integrand11[i] = a.dot(b1[:, i])
        integrand12[i] = a.dot(b2[:, i])
        integrand22[i] = b.dot(b2[:, i])

    j = zeros((2, 2))
    j[0, 0] = trapz(integrand11, linspace(0, 1, T))
    j[0, 1] = trapz(integrand12, linspace(0, 1, T))
    j[1, 1] = trapz(integrand22, linspace(0, 1, T))
    j[1, 0] = j[0, 1]

    return (j)


def shift_f(f, tau):
    """
    shifts a curve f by tau

    :param f: numpy ndarray of shape (2,M) of M samples
    :param tau: scalar

    :rtype: numpy ndarray
    :return fn: shifted curve

    """
    n, T = f.shape
    fn = zeros((n, T))
    fn[:, 0:(T - 1)] = roll(f[:, 0:(T - 1)], tau, axis=1)
    fn[:, T - 1] = fn[:, 0]

    return (fn)


def find_rotation_and_seed_coord(beta1, beta2):
    """
    This function returns a candidate list of optimally oriented and
    registered (seed) shapes w.r.t. beta1

    :param beta1: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return beta2new: optimal rotated beta2 to beta1
    :return O: rotation matrix
    :return tau: seed

    """
    n, T = beta1.shape
    q1 = curve_to_q(beta1)
    Ltwo = zeros(T)
    Rlist = zeros((n, n, T))
    for ctr in range(0, T):
        beta2n = shift_f(beta2, ctr)
        beta2new, R = find_best_rotation(beta1, beta2n)
        q2new = curve_to_q(beta2new)
        Ltwo[ctr] = innerprod_q2(q1 - q2new, q1 - q2new)
        Rlist[:, :, ctr] = R

    tau = Ltwo.argmin()
    O_hat = Rlist[:, :, tau]
    beta2new = shift_f(beta2, tau)
    beta2new = O_hat.dot(beta2new)

    return (beta2new, O_hat, tau)


def find_rotation_and_seed_q(q1, q2):
    """
    This function returns a candidate list of optimally oriented and
    registered (seed) shapes w.r.t. beta1

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return beta2new: optimal rotated beta2 to beta1
    :return O: rotation matrix
    :return tau: seed

    """
    n, T = q1.shape
    Ltwo = zeros(T)
    Rlist = zeros((n, n, T))
    for ctr in range(0, T):
        q2n = shift_f(q2, ctr)
        q2new, R = find_best_rotation(q1, q2n)
        Ltwo[ctr] = innerprod_q2(q1 - q2new, q1 - q2new)
        Rlist[:, :, ctr] = R

    tau = Ltwo.argmin()
    O_hat = Rlist[:, :, tau]
    q2new = shift_f(q2, tau)
    q2new = O_hat.dot(q2new)

    return (q2new, O_hat, tau)


def group_action_by_gamma_coord(f, gamma):
    """
    This function reparamerized curve f by gamma

    :param f: numpy ndarray of shape (2,M) of M samples
    :param gamma: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return fn: reparatermized curve

    """
    n, T = f.shape
    fn = zeros((n, T))

    for j in range(0, n):
        s = interp1d(linspace(0, 1, T), f[j, :])
        fn[j, :] = s(gamma)

    return (fn)


def group_action_by_gamma(q, gamma):
    """
    This function reparamerized srvf q by gamma

    :param f: numpy ndarray of shape (2,M) of M samples
    :param gamma: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return qn: reparatermized srvf

    """
    n, T = q.shape
    gammadot = gradient(gamma, 1. / T)
    qn = zeros((n, T))

    for j in range(0, n):
        s = InterpolatedUnivariateSpline(linspace(0, 1, T), q[j, :], k=3)
        qn[j, :] = s(gamma) * sqrt(gammadot)

    qn = qn / sqrt(innerprod_q2(qn, qn))

    return (qn)


def project_curve(q):
    """
    This function projects srvf q to set of close curves

    :param q: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return qproj: project srvf
    """
    n,T = q.shape
    if n==2:
        dt = 0.35
    if n==3:
        dt = 0.2
    epsilon = 1e-6

    iter = 1
    res = ones(n)
    J = zeros((n,n))

    s = linspace(0,1,T)

    qnew = q.copy()
    qnew = qnew / sqrt(innerprod_q2(qnew,qnew))

    qnorm = zeros(T)
    G = zeros(n)
    C = zeros(300)
    while (norm(res) > epsilon):
        if iter > 300:
            break

        # Jacobian
        for i in range(0,n):
            for j in range(0,n):
                J[i,j] = 3 * trapz(qnew[i,:]*qnew[j,:],s)
        
        J += eye(n)

        for i in range(0,T):
            qnorm[i] = norm(qnew[:,i])
        
        # Compute the residue
        for i in range(0,n):
            G[i] = trapz(qnew[i,:]*qnorm,s)
        
        res = -G

        if (norm(res) < epsilon):
            break

        x = solve(J,res)
        C[iter] = norm(res)

        delG = Basis_Normal_A(qnew)
        temp = 0
        for i in range(0,n):
            temp += x[i]*delG[i]*dt
        
        qnew += temp
        iter += 1
    
    qnew = qnew/sqrt(innerprod_q2(qnew,qnew))

    return qnew


def pre_proc_curve(beta, T=100):
    """
    This function prepcoessed a curve beta to set of closed curves

    :param beta: numpy ndarray of shape (2,M) of M samples
    :param T: number of samples (default = 100)

    :rtype: numpy ndarray
    :return betanew: projected beta
    :return qnew: projected srvf
    :return A: alignment matrix (not used currently)

    """
    beta = resamplecurve(beta, T)
    q = curve_to_q(beta)
    qnew = project_curve(q)
    x = q_to_curve(qnew)
    a = -1 * calculatecentroid(x)
    betanew = x + tile(a, [T, 1]).T
    A = eye(2)

    return (betanew, qnew, A)


def elastic_distance_curve(beta1, beta2):
    """
    Calculates the two elastic distances between two curves
    :param beta1: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples

    :rtype: scalar
    :return dist: distance
    """

    if (beta1 == beta2).all():
        d = 0.0
    else:
        v,d = inverse_exp_coord(beta1, beta2)

    return d
    

def inverse_exp_coord(beta1, beta2):
    """
    Calculate the inverse exponential to obtain a shooting vector from
    beta1 to beta2 in shape space of open curves

    :param beta1: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return v: shooting vectors
    :return dist: distance

    """
    T = beta1.shape[1]
    centroid1 = calculatecentroid(beta1)
    beta1 = beta1 - tile(centroid1, [T, 1]).T
    centroid2 = calculatecentroid(beta2)
    beta2 = beta2 - tile(centroid2, [T, 1]).T

    q1 = curve_to_q(beta1)

    # Iteratively optimize over SO(n) x Gamma
    for i in range(0, 1):
        # Optimize over SO(n)
        beta2, O_hat, tau = find_rotation_and_seed_coord(beta1, beta2)
        q2 = curve_to_q(beta2)

        # Optimize over Gamma
        gam = optimum_reparam_curve(q2, q1, 0.0)
        gamI = uf.invertGamma(gam)
        # Applying optimal re-parameterization to the second curve
        beta2 = group_action_by_gamma_coord(beta2, gamI)

    # Optimize over SO(n)
    beta2, O_hat, tau = find_rotation_and_seed_coord(beta1, beta2)
    q2n = curve_to_q(beta2)

    # Compute geodesic distance
    q1dotq2 = innerprod_q2(q1, q2n)
    dist = arccos(q1dotq2)

    # Compute shooting vector
    if q1dotq2 > 1:
        q1dotq2 = 1

    u = q2n - q1dotq2 * q1
    normu = sqrt(innerprod_q2(u, u))

    if normu > 1e-4:
        v = u * arccos(q1dotq2) / normu
    else:
        v = zeros((2, T))

    return (v, dist)


def inverse_exp(q1, q2, beta2):
    """
    Calculate the inverse exponential to obtain a shooting vector from
    q1 to q2 in shape space of open curves

    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples
    :param beta2: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return v: shooting vectors

    """
    T = q1.shape[1]
    centroid1 = calculatecentroid(beta2)
    beta2 = beta2 - tile(centroid1, [T, 1]).T

    # Optimize over SO(n)
    q2, O_hat, tau = find_rotation_and_seed_q(q1, q2)

    # Optimize over Gamma
    gam = optimum_reparam_curve(q2, q1, 0.0)
    gamI = uf.invertGamma(gam)

    # Applying optimal re-parameterization to the second curve
    beta2 = group_action_by_gamma_coord(beta2, gamI)
    q2 = curve_to_q(beta2)

    # Optimize over SO(n)
    q2, O2, tau = find_rotation_and_seed_q(q1, q2)

    # Compute geodesic distance
    q1dotq2 = innerprod_q2(q1, q2)
    dist = arccos(q1dotq2)

    # Compute shooting vector
    if q1dotq2 > 1:
        q1dotq2 = 1

    u = q2 - q1dotq2 * q1
    normu = sqrt(innerprod_q2(u, u))

    if normu > 1e-4:
        v = u * arccos(q1dotq2) / normu
    else:
        v = zeros((2, T))

    return v


def gram_schmidt(basis):
    """
   Performs Gram Schmidt Orthogonlization of a basis_o

    :param basis: list of numpy ndarray of shape (2,M) of M samples

    :rtype: list of numpy ndarray
    :return basis_o: orthogonlized basis

    """
    b1 = basis[0]
    b2 = basis[1]

    basis1 = b1 / sqrt(innerprod_q2(b1, b1))
    b2 = b2 - innerprod_q2(basis1, b2) * basis1
    basis2 = b2 / sqrt(innerprod_q2(b2, b2))

    basis_o = [basis1, basis2]

    return (basis_o)


def project_tangent(w, q, basis):
    """
    projects srvf to tangent space w using basis

    :param w: numpy ndarray of shape (2,M) of M samples
    :param q: numpy ndarray of shape (2,M) of M samples
    :param basis: list of numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return wproj: projected q

    """
    w = w - innerprod_q2(w, q) * q
    bo = gram_schmidt(basis)

    wproj = w - innerprod_q2(w, bo[0]) * bo[0] - innerprod_q2(w, bo[1]) * bo[1]

    return (wproj)


def scale_curve(beta):
    """
    scales curve to length 1

    :param beta: numpy ndarray of shape (2,M) of M samples

    :rtype: numpy ndarray
    :return beta_scaled: scaled curve
    :return scale: scale factor used

    """
    n, T = beta.shape
    normbetadot = zeros(T)
    betadot = gradient(beta, 1. / T)
    betadot = betadot[1]
    for i in range(0, T):
        normbetadot[i] = norm(betadot[:, i])

    scale = trapz(normbetadot, linspace(0, 1, T))
    beta_scaled = beta / scale

    return (beta_scaled, scale)


def parallel_translate(w, q1, q2, basis, mode=0):
    """
    parallel translates q1 and q2 along manifold

    :param w: numpy ndarray of shape (2,M) of M samples
    :param q1: numpy ndarray of shape (2,M) of M samples
    :param q2: numpy ndarray of shape (2,M) of M samples
    :param basis: list of numpy ndarray of shape (2,M) of M samples
    :param mode: open 0 or closed curves 1 (default 0)

    :rtype: numpy ndarray
    :return wbar: translated vector

    """
    modes = [0, 1]
    mode = [i for i, x in enumerate(modes) if x == mode]
    if len(mode) == 0:
        mode = 0
    else:
        mode = mode[0]

    wtilde = w - 2 * innerprod_q2(w, q2) / innerprod_q2(q1 + q2, q1 + q2) * (q1 + q2)
    l = sqrt(innerprod_q2(wtilde, wtilde))

    if mode == 1:
        wbar = project_tangent(wtilde, q2, basis)
        normwbar = sqrt(innerprod_q2(wbar, wbar))
        if normwbar > 10 ** (-4):
            wbar = wbar * l / normwbar
    else:
        wbar = wtilde

    return (wbar)


def curve_zero_crossing(Y, beta, bt, y_max, y_min, gmax, gmin):
    """
    finds zero-crossing of optimal gamma, gam = s*gmax + (1-s)*gmin
    from elastic curve regression model

    :param Y: response
    :param beta: predicitve function
    :param bt: basis function
    :param y_max: maximum repsonse for warping function gmax
    :param y_min: minimum response for warping function gmin
    :param gmax: max warping function
    :param gmin: min warping fucntion

    :rtype: numpy array
    :return gamma: optimal warping function
    :return O_hat: rotation matrix

    """
    # simple iterative method based on intermediate theorem
    T = beta.shape[1]
    betanu = q_to_curve(bt)
    max_itr = 100
    a = zeros(max_itr)
    a[0] = 1
    f = zeros(max_itr)
    f[0] = y_max - Y
    f[1] = y_min - Y
    mrp = f[0]
    mrn = f[1]
    mrp_ind = 0  # most recent positive index
    mrn_ind = 1  # most recent negative index

    for ii in range(2, max_itr):
        x1 = a[mrp_ind]
        x2 = a[mrn_ind]
        y1 = mrp
        y2 = mrn
        a[ii] = (x1 * y2 - x2 * y1) / float(y2 - y1)

        beta1, O_hat, tau = find_rotation_and_seed_coord(betanu, beta)

        gamma = a[ii] * gmax + (1 - a[ii]) * gmin

        beta1 = group_action_by_gamma_coord(beta1, gamma)
        beta1, O_hat1, tau = find_rotation_and_seed_coord(betanu, beta1)
        q1 = curve_to_q(beta1)
        f[ii] = innerprod_q2(q1, bt) - Y

        if fabs(f[ii]) < 1e-5:
            break
        elif f[ii] > 0:
            mrp = f[ii]
            mrp_ind = ii
        else:
            mrn = f[ii]
            mrn_ind = ii

    gamma = a[ii] * gmax + (1 - a[ii]) * gmin

    beta1, O_hat, tau = find_rotation_and_seed_coord(betanu, beta)
    beta1 = group_action_by_gamma_coord(beta1, gamma)
    beta1, O_hat1, tau = find_rotation_and_seed_coord(betanu, beta1)
    O_hat = O_hat.dot(O_hat1)

    return (gamma, O_hat, tau)


def elastic_shooting(q1,v):
    """
    Calculates shooting vector from v to q1

    :param q1: vector of srvf
    :param v: shooting vector

    :rtype numpy ndarray
    :return q2n: vector of srvf
    """
    d = sqrt(innerprod_q2(v,v))
    if d < 0.00001:
        q2n = q1
    else:
        q2n = cos(d)*q1 + (sin(d)/d)*v
        q2n = project_curve(q2n)
    
    return (q2n)


def rot_mat(theta):
    O = array([(cos(theta), -1*sin(theta)), (sin(theta), cos(theta))])

    return (O)



cimport cmlogit
import numpy as np
from scipy.linalg import norm

cimport numpy as np
from cpython cimport array


def mlogit_warp(np.ndarray[double, ndim=1, mode="c"] alpha,
                np.ndarray[double, ndim=2, mode="c"] beta,
                np.ndarray[double, ndim=1, mode="c"] time,
                np.ndarray[double, ndim=1, mode="c"] q,
                np.ndarray[int, ndim=1, mode="c"] y,
                max_iter=4000, tol=1e-10, delta=0.008, display=0):
    """
    cython interface perform warping calculation for multinomial cost function

    :param alpha: vector of size m:number of classes
    :param beta: matrix of size Nxm
    :param time: vector of size N describing the sample points
    :param q: numpy vector of size M srsf
    :param y: numpy ndarray of shape m class labels
    :param max_iter: maximal number of iterations (default = 400)
    :param tol: stopping tolerance (default = 1e-4)
    :param delta: step size (default = 0.008)
    :param display: show iterations (default = 0)

    :rtype numpy ndarray
    :return gamo: describing the warping function

    """
    cdef int m1, m, max_itri, displayi
    cdef double toli, deltai
    toli = tol
    deltai = delta
    max_itri = max_iter
    displayi = display
    m1 = time.size
    m = beta.shape[1]
    alpha = alpha/norm(alpha)
    q = q/norm(q)
    for i in range(0, m):
        beta[:, i] = beta[:, i]/norm(beta[:, i])

    cdef np.ndarray[double, ndim = 1, mode = "c"] gam1 = np.linspace(0, 1, m1)
    cdef np.ndarray[double, ndim = 1, mode = "c"] beta1 = np.zeros(m1 * m)
    cdef np.ndarray[double, ndim = 1, mode = "c"] gamout = np.zeros(m1)

    for ii in xrange(0, m):
        beta1[ii * m1:ii * m1 + m1] = beta[:, ii]


    gam1 = np.ascontiguousarray(gam1)
    beta1 = np.ascontiguousarray(beta1)
    gamout = np.ascontiguousarray(gamout)

    cmlogit.mlogit_warp_grad(&m1, &m, &alpha[0], &beta1[0], &time[0], &gam1[0], &q[0], &y[0], &max_itri, &toli, &deltai, &displayi, &gamout[0])

    return gamout
